        subroutine input(maxnode,maxcell,maxedge, nnode,ncell,nedge,
     &                   x, q,cell, edge,ecell,boun,pert1,pert3)
        implicit none
c
        integer*4 maxnode,maxcell,maxedge, nnode,ncell,nedge,
     &            cell(4,maxcell), edge(2,maxedge),
     &            ecell(2,maxedge), boun(maxedge)
        real*8    x(2,maxnode), q(4,maxcell),
     &            pert1(2,maxnode), pert3(2,maxnode)
c
#include "const.inc"
c
        integer in,ie,ic,ipde
        real*8  p,r,u,e, xt,yt
c
c------ set universal constants
c
        gam = 1.4d0
        gm1 = gam - 1.0d0
        cfl = 0.9d0
        eps = 0.05d0
c
c------ read in data from grid file
c
        open(1,file='grid.dat',status='old')
c
        read(1,*) nnode, ncell, nedge
        ncell = ncell+1  ! far-field dummy cell
        if(nnode .gt. maxnode) stop 'maxnode too small'
        if(ncell .gt. maxcell) stop 'maxcell too small'
        if(nedge .gt. maxedge) stop 'maxedge too small'
c
        do in = 1, nnode
          read(1,*) x(1,in),x(2,in)
        enddo
c
        do ic = 1, ncell-1
          read(1,*) cell(1,ic),cell(2,ic),cell(3,ic),cell(4,ic)
        enddo
c
        do ie = 1, nedge
          read(1,*) edge(1,ie),edge(2,ie),
     &              ecell(1,ie),ecell(2,ie),boun(ie)
          if(boun(ie).eq.2) then
           ecell(2,ie) = ncell
            boun(ie) = 0
          endif
        enddo
c
        close(1)
c
c------ read in data from flow file, initialising if necessary
c
        open(1,file='flow.dat',status='old')
        read(1,*) p, r, mach, coef(1),coef(2), coef(3)
        coef(2) = coef(2)*datan(1.0d0)/45.0d0
        p = 1.0d0
        r = 1.0d0
        u = dsqrt(gam*p/r)*mach
        e = p/(r*gm1) + 0.5d0*u**2
c
        do ic = 1, ncell
          q(1,ic) = r
          q(2,ic) = r*u
          q(3,ic) = 0.d0
          q(4,ic) = r*e
        enddo
c
        do ic = 1, ncell
          read(1,*,err=999,end=999) (q(ipde,ic),ipde=1,4)
        enddo
 999    close(1)
c
        open(1,file='grid_pert1.dat',status='old')
        do in = 1,nnode
          read(1,*) pert1(1,in), pert1(2,in)
        enddo
        close(1)
c
        open(1,file='grid_pert3.dat',status='old')
        do in = 1,nnode
          read(1,*) pert3(1,in), pert3(2,in)
        enddo
        close(1)
c
        return
        end


        subroutine output(maxcell,ncell, q)
        implicit none
c
        integer*4 ncell, ic, ipde, maxcell
        real*8    q(4,maxcell), p, r, mach,coef(3)
c
        open(1,file='flow.dat',status='old')
        read(1,*) p, r, mach, coef(1), coef(2), coef(3)
        close(1)
c
        open(1,file='flow.dat',status='old')
        write(1,100) p, r, mach, coef(1), coef(2), coef(3)
        do ic = 1, ncell-1
          write(1,100) (q(ipde,ic),ipde=1,4)
        enddo
 100    format(e23.16,' ',e23.16,' ',e23.16,' ',e23.16)
        close(1)
c
        return
        end

        subroutine output_lin(maxcell,ncell, qd)
        implicit none
c
        integer*4 ncell, ic, ipde, maxcell
        real*8    qd(4,maxcell), p, r, mach,coef(3)
c
        open(1,file='flow.dat',status='old')
        read(1,*) p, r, mach, coef(1), coef(2), coef(3)
        close(1)
c
        open(1,file='flow_lin.dat',status='new')
        write(1,100) p, r, mach, coef(1), coef(2), coef(3)
        do ic = 1, ncell-1
          write(1,100) (qd(ipde,ic),ipde=1,4)
        enddo
        close(1)
 100    format(e23.16,' ',e23.16,' ',e23.16,' ',e23.16)
c
        return
        end
c
        subroutine output_adj(maxcell,ncell, qb)
        implicit none
c
        integer*4 ncell, ic, ipde,maxcell
        real*8    qb(4,maxcell), p, r, mach,coef(3)
c
        open(1,file='flow.dat',status='old')
        read(1,*) p, r, mach, coef(1), coef(2), coef(3)
        close(1)
c
        open(1,file='flow_adj.dat',status='new')
        write(1,100) p, r, mach, coef(1), coef(2), coef(3)
        do ic = 1, ncell-1
          write(1,100) (qb(ipde,ic),ipde=1,4)
        enddo
        close(1)
 100    format(e23.16,' ',e23.16,' ',e23.16,' ',e23.16)
c
        return
        end
c
        subroutine input_hes(maxcell,ncell,qd1,qd2,qd3,qb)
     &                       
        implicit none
c
        integer*4 ncell,ic,ipde,maxcell
        real*8 qd1(4,maxcell), qb(4,maxcell),
     &         p, r, mach, coef(3),
     &         qd2(4,maxcell),qd3(4,maxcell)
c
        open(1,file='flow_lin1.dat',status='old')
        read(1,*) p, r, mach, coef(1), coef(2), coef(3)
        do ic=1, ncell-1
          read(1,*) (qd1(ipde,ic),ipde=1,4)
        enddo
        close(1)
c
        open(1,file='flow_lin2.dat',status='old')
        read(1,*) p, r, mach, coef(1), coef(2), coef(3)
        do ic=1, ncell-1
          read(1,*) (qd2(ipde,ic),ipde=1,4)
        enddo
        close(1)
c
        open(1,file='flow_lin3.dat',status='old')
        read(1,*) p, r, mach, coef(1), coef(2), coef(3)
        do ic=1, ncell-1
          read(1,*) (qd3(ipde,ic),ipde=1,4)
        enddo
        close(1)
c
        do ic=1,4
          qd1(ic,ncell) = 0.0d0
          qd2(ic,ncell) = 0.0d0
          qd3(ic,ncell) = 0.0d0
        enddo
c
        open(1,file='flow_adj.dat',status='old')
        read(1,*) p, r, mach, coef(1), coef(2), coef(3)
        do ic=1, ncell-1
          read(1,*) (qb(ipde,ic),ipde=1,4)
        enddo
        close(1)
c
        qb(1,ncell) = 0.0d0
        qb(2,ncell) = 0.0d0
        qb(3,ncell) = 0.0d0
        qb(4,ncell) = 0.0d0
c
        return
        end
c
